
import os
import pickle
import rawpy
import time
import cv2
import numpy as np
import matplotlib.pyplot as plt
import tensorflow as tf
from pprint import pprint
from srccode.util import *
from scipy import interpolate, ndimage, signal

def getImgRef(imgPath):

  dataset = {}

  inputList = os.listdir(imgPath)

  for img in inputList:
    ref = os.path.join(imgPath, img, 'reference_frame.txt')
    with open(ref) as f:
      tag = f.read()
    dataset[img] = tag.strip()

  return dataset

def savePickle(data, fileName):
  with open(fileName, 'wb') as f:
    pickle.dump(data, f)

def loadPickle(fileName):
  with open(fileName, 'rb') as f:
    loadedData = pickle.load(f)
  return loadedData

def imgPrep(PROJ_PATH, dataDir):

  dataPath = os.path.join(PROJ_PATH, 'data', dataDir)
  imgPath  = os.path.join(dataPath, 'results_20171023')
  dataset  = getImgRef(imgPath)

  for imgName, tag in dataset.items():

    path    = os.path.join(dataPath, f'bursts/{imgName}/payload_N00{tag}.dng')
    hdrPath = os.path.join(dataPath, f'results_20171023/{imgName}/final.jpg')

    print(imgName, path)

    with rawpy.imread(path) as raw:
      img_raw = raw.raw_image_visible

    img = img_raw / 255
    plt.imshow(img, cmap="gray", vmin=0, vmax=1)
    plt.show()

    time.sleep(1)
    k = input()
    if int(k) > 0 :
      print("Rotating image...")
      img = np.rot90(img, k=int(k)).copy()
      plt.imshow(img, cmap="gray", vmin=0, vmax=1)
      plt.show()

    img_bgr = cv2.imread(hdrPath)
    imgGt = cv2.cvtColor(img_bgr, cv2.COLOR_BGR2RGB)/255
    plt.imshow(imgGt)
    plt.show()

    rawPath = PROJ_PATH+f'data/train/raw/{imgName}.raw'
    gtPath  = PROJ_PATH+f'data/train/gt/{imgName}.gt'

    ans = input()
    if ans == 'y':
      print("Saving image...")
      savePickle(img, rawPath)
      savePickle(imgGt, gtPath)

class dataGenerator(tf.keras.utils.Sequence):

  def __init__(self, dataPath, numBatches, numPatches, patchSize):
    self.DATA_PATH   = dataPath
    self.RAW_LIST    = os.listdir(os.path.join(self.DATA_PATH, 'raw')) 
    self.GT_LIST     = os.listdir(os.path.join(self.DATA_PATH, 'gt')) 
    self.NUM_IMGS    = len(self.RAW_LIST)
    self.NUM_BATCHES = numBatches
    self.NUM_PATCHES = numPatches 
    self.SIZE_PATCH  = patchSize

  def __len__(self):
    return self.NUM_BATCHES if self.NUM_IMGS > self.NUM_BATCHES else self.NUM_IMGS

  def __getitem__(self, index):  
    rawName  = self.RAW_LIST[index]
    gtName   = self.GT_LIST[index]
    
    imageRaw = loadPickle(os.path.join(self.DATA_PATH, 'raw', rawName))
    imageGt  = loadPickle(os.path.join(self.DATA_PATH, 'gt', gtName))
    H, W, _  = imageGt.shape

    maskR, maskG, maskB = genRGBMask(imageRaw.shape)

    imageRawR = (imageRaw * maskR)
    imageRawG = (imageRaw * maskG)
    imageRawB = (imageRaw * maskB)

    yPos = np.random.choice(np.arange(H-self.SIZE_PATCH[0]-1, step = self.SIZE_PATCH[0]), size=self.NUM_PATCHES)
    xPos = np.random.choice(np.arange(W-self.SIZE_PATCH[1]-1, step = self.SIZE_PATCH[1]), size=self.NUM_PATCHES) 

    rawBatch = []
    gtBatch  = []  
    for i in range(self.NUM_PATCHES):
      
      rawPatch = np.stack((imageRawR[yPos[i]:yPos[i]+self.SIZE_PATCH[0], 
                                     xPos[i]:xPos[i]+self.SIZE_PATCH[1]], 
                           imageRawG[yPos[i]:yPos[i]+self.SIZE_PATCH[0], 
                                     xPos[i]:xPos[i]+self.SIZE_PATCH[1]], 
                           imageRawB[yPos[i]:yPos[i]+self.SIZE_PATCH[0], 
                                     xPos[i]:xPos[i]+self.SIZE_PATCH[1]]), 
                          axis=2)
      

      gtPatch  = imageGt[yPos[i]:yPos[i]+self.SIZE_PATCH[0], 
                         xPos[i]:xPos[i]+self.SIZE_PATCH[1], 
                         ...]
      
      rawBatch.append(rawPatch)
      gtBatch.append(gtPatch)     

    X = np.stack(rawBatch, axis=0)
    Y = np.stack(gtBatch, axis=0)

    return X, Y

def demo(model, PROJ_PATH):

  imageList = os.listdir(PROJ_PATH+f'data/demo/raw/') 
  idx       = np.random.randint(len(imageList), size=1)[0]
  idx       = 1
  imgName   = os.path.splitext(imageList[idx])[0]
  rawPath   = PROJ_PATH+f'data/demo/raw/{imgName}.raw'
  rawPathP  = PROJ_PATH+f'data/demo/final/payload_N000.dng'
  gtPath    = PROJ_PATH+f'data/demo/gt/{imgName}.gt'

  imageRaw  = loadPickle(rawPath)
  imageGt   = loadPickle(gtPath)

  with rawpy.imread(rawPathP) as raw:
    #img_raw = raw.raw_image_visible
    img     = raw.postprocess()

  H,W       = imageRaw.shape
  maskR, maskG, maskB = genRGBMask((H,W))
  imageRawR = imageRaw * maskR
  imageRawG = imageRaw * maskG
  imageRawB = imageRaw * maskB

  imageRawRGB = np.stack((imageRawR, imageRawG, imageRawB), axis=2)
  
  imageRGBQ0 = model(tf.expand_dims(imageRawRGB[0:int(H/2), 0:int(W/2), ...], axis=0))
  imageRGBQ1 = model(tf.expand_dims(imageRawRGB[0:int(H/2), int(W/2):W, ...], axis=0))
  imageRGBQ2 = model(tf.expand_dims(imageRawRGB[int(H/2):H, 0:int(W/2), ...], axis=0))
  imageRGBQ3 = model(tf.expand_dims(imageRawRGB[int(H/2):H, int(W/2):W, ...], axis=0))

  Q0Shape    = imageRGBQ0.shape
  Q1Shape    = imageRGBQ1.shape
  Q2Shape    = imageRGBQ2.shape
  Q3Shape    = imageRGBQ3.shape

  imageRGBL  = np.concatenate((imageRGBQ0[:, 0:Q0Shape[1], :, ...], 
                               imageRGBQ2[:, 0:Q2Shape[1],   :, ...]), axis=1)
  imageRGBR  = np.concatenate((imageRGBQ1[:, 0:Q1Shape[1], :, ...], 
                               imageRGBQ3[:, 0:Q3Shape[1],   :, ...]), axis=1)
  lShape     = imageRGBL.shape
  rShape     = imageRGBR.shape

  imageRGB      = np.concatenate((imageRGBL[:, :, 0:lShape[2]-2, ...], 
                                  imageRGBR[:, :, 2:rShape[2], ...]), axis=2)
  
  #imageRGB = model(tf.expand_dims(imageRawRGB[:, :, ...], axis=0))
  ssim   = tf.image.ssim(imageGt, np.squeeze(imageRGB, axis=0).astype(imageGt.dtype), max_val=1)
  psnr   = tf.image.psnr(imageGt, np.squeeze(imageRGB, axis=0).astype(imageGt.dtype), max_val=1)

  print(f"SSIM: {ssim}, PSNR: {psnr}")

  fig, ax = plt.subplots(nrows=2, ncols = 2, figsize=(40, 40))

  ax[0,0].imshow(imageRaw, cmap="gray")
  ax[0,0].axis('off')
  ax[0,0].set_title(f'Raw Image', size=20)

  ax[0,1].imshow(img)
  ax[0,1].axis('off')
  ax[0,1].set_title(f'Post-processed by rawPy', size=20)

  ax[1,0].imshow(imageGt)
  ax[1,0].axis('off')
  ax[1,0].set_title(f'Gound Truth Image (HDR+)', size=20)

  ax[1,1].imshow(np.squeeze(imageRGB, axis=0))
  ax[1,1].axis('off')
  ax[1,1].set_title(f'Reconstructed Image', size=20)

  fig.subplots_adjust(wspace=0, hspace = 0.1)
  fig.show()
  #fig.savefig('image_compare.png', bbox_inches='tight')

  return

def gammaCorrection(img):

    gamma = 2.2
    img_corrected = np.clip(img, a_min=0, a_max=1) ** (1/gamma)
    
    return img_corrected

def hq_linear(img, imgSize):

    filter0 = np.array([[ 0, 0, -1, 0,  0],
                        [ 0, 0,  2, 0,  0],
                        [-1, 2,  4, 2, -1],
                        [ 0, 0,  2, 0,  0],
                        [ 0, 0, -1, 0,  0]])
    filter0 = filter0 / np.sum(filter0)

    filter1 = np.array([[ 0,  0, 1/2,  0,  0],
                        [ 0, -1,   0, -1,  0],
                        [-1,  4,   5,  4, -1],
                        [ 0, -1,   0, -1,  0],
                        [ 0, 0,  1/2,  0,  0]])
    filter1 = filter1 / np.sum(filter1)

    filter2 = np.array([[  0,  0, -1,  0,   0],
                        [  0, -1,  4, -1,   0],
                        [1/2,  0,  5,  0, 1/2],
                        [  0, -1,  4, -1,   0],
                        [  0,  0, -1,  0,   0]])
    filter2 = filter2 / np.sum(filter2)
    
    filter3 = np.array([[   0,  0, -3/2,  0,    0],
                        [   0,  2,    0,  2,    0],
                        [-3/2,  0,    6,  0, -3/2],
                        [   0,  2,    0,  2,    0],
                        [   0,  0, -3/2,  0,    0]])
    filter3 = filter3 / np.sum(filter3)
    
    img = img*255

    channel0 = np.clip(signal.convolve2d(img, filter0, mode='same'), a_min=0, a_max=255)
    channel1 = np.clip(signal.convolve2d(img, filter1, mode='same'), a_min=0, a_max=255)
    channel2 = np.clip(signal.convolve2d(img, filter2, mode='same'), a_min=0, a_max=255)
    channel3 = np.clip(signal.convolve2d(img, filter3, mode='same'), a_min=0, a_max=255)

    # Red channel interpolation
    xr = np.arange(0, imgSize[1], 2)
    yr = np.arange(0, imgSize[0], 2)

    Xr, Yr = np.meshgrid(xr, yr)

    # Blue channel interpolation
    xb = np.arange(1, imgSize[1], 2)
    yb = np.arange(1, imgSize[0], 2)

    Xb, Yb = np.meshgrid(xb, yb)

    # R row, B column Coordinates
    x_rb = np.arange(1, imgSize[1], 2)
    y_rb = np.arange(0, imgSize[0], 2)

    X_rb, Y_rb = np.meshgrid(x_rb, y_rb)

    # B row, R column Coordinates
    x_br = np.arange(0, imgSize[1], 2)
    y_br = np.arange(1, imgSize[0], 2)

    X_br, Y_br = np.meshgrid(x_br, y_br)

    # Red channel
    imgR             = np.copy(img)
    imgR[Y_rb, X_rb] = channel1[Y_rb, X_rb]
    imgR[Y_br, X_br] = channel2[Y_br, X_br]    
    imgR[Yb, Xb]     = channel3[Yb, Xb]    
    
    # Green channel
    imgG             = np.copy(img)
    imgG[Yr, Xr]     = channel0[Yr, Xr]
    imgG[Yb, Xb]     = channel0[Yb, Xb]
    
    # Blue channel
    imgB             = np.copy(img)
    imgB[Y_br, X_br] = channel1[Y_br, X_br]    
    imgB[Y_rb, X_rb] = channel2[Y_rb, X_rb]
    imgB[Yr, Xr]     = channel3[Yr, Xr]    

    img_final = np.stack([imgR, imgG, imgB], axis=2)

    img_gamma = gammaCorrection(img_final/255)
    img_gamma = (img_gamma * 255).astype(np.uint8)

    return img_gamma


def linear(img, imgSize):
    img = img*255

    x = np.arange(0, imgSize[1])
    y = np.arange(0, imgSize[0])

    # Q1
    
    # Red channel interpolation
    xr = np.arange(0, imgSize[1], 2)
    yr = np.arange(0, imgSize[0], 2)

    Xr, Yr = np.meshgrid(xr, yr)
    
    zr = np.copy(img[Yr, Xr])

    fr = interpolate.interp2d(xr, yr, zr)

    imgR = fr(x, y)

    # Blue channel interpolation
    xb = np.arange(1, imgSize[1], 2)
    yb = np.arange(1, imgSize[0], 2)

    Xb, Yb = np.meshgrid(xb, yb)
    
    zb = np.copy(img[Yb, Xb])

    fb = interpolate.interp2d(xb, yb, zb)

    imgB = fb(x, y)

    # Green channel interpolation

    g0n1 = np.roll(img, -1, axis=0)
    g0p1 = np.roll(img,  1, axis=0)
    g1n1 = np.roll(img, -1, axis=1)
    g1p1 = np.roll(img,  1, axis=1)

    fg   = np.average(np.array([g0n1, g0p1, g1n1, g1p1]), axis=0)

    imgG = np.copy(img)
    imgG[Yr, Xr] = fg[Yr, Xr]
    imgG[Yb, Xb] = fg[Yb, Xb]

    # construct the final image
    img_final = np.stack([imgR, imgG, imgB], axis=2)

    img_gamma = gammaCorrection(img_final/255)

    img_gamma = (img_gamma * 255).astype(np.uint8)

    return img_gamma



