import math
import tensorflow as tf
from srccode.data import *

class HdrNet(tf.keras.Model):
    def __init__(self, params):
      super().__init__()

      # Output channel options
      self.params         = params
      self.initializer    = tf.initializers.VarianceScaling(scale=params['INIT_SCALE'])
      self.regularizer    = tf.keras.regularizers.L2(params['L2_REG_WEIGHT'])
      self.dropout_rate   = params['DROPOUT_RATE']
      self.inputLayer     = tf.keras.layers.InputLayer(input_shape=(None, None, 3))
      self.filterBase     = 64

      self.downSampleLayer_128_3_BN_0 = tf.keras.Sequential([               
               tf.keras.layers.Conv2D(filters            = self.filterBase * 2, 
                                      kernel_size        = 3, 
                                      strides            = (2,2),
                                      padding            = 'same',
                                      kernel_regularizer = self.regularizer, 
                                      kernel_initializer = self.initializer),
               tf.keras.layers.LeakyReLU(),   
               tf.keras.layers.AveragePooling2D(pool_size=(2, 2), 
                         strides=(1, 1), padding='same')

      ], name="downSampleLayer_128_3_BN_0")
      
      self.downSampleLayer_256_3_BN_0 = tf.keras.Sequential([
               tf.keras.layers.Conv2D(filters            = self.filterBase * 4, 
                                      kernel_size        = 3, 
                                      strides            = (2,2),
                                      padding            = 'same',
                                      kernel_regularizer = self.regularizer, 
                                      kernel_initializer = self.initializer),
               tf.keras.layers.BatchNormalization(),
               tf.keras.layers.LeakyReLU(),          
               tf.keras.layers.AveragePooling2D(pool_size=(2, 2), 
                         strides=(1, 1), padding='same'),  
               tf.keras.layers.Conv2D(filters            = self.filterBase * 4, 
                                      kernel_size        = 3, 
                                      strides            = (1,1),
                                      padding            = 'same',
                                      kernel_regularizer = self.regularizer, 
                                      kernel_initializer = self.initializer), 
               tf.keras.layers.BatchNormalization(),
               tf.keras.layers.LeakyReLU(),                                                                          
      ], name="downSampleLayer_256_3_BN_0")
      
      self.downSampleLayer_256_3_BN_1 = tf.keras.Sequential([
               tf.keras.layers.Conv2D(filters            = self.filterBase * 4, 
                                      kernel_size        = 3, 
                                      strides            = (2,2),
                                      padding            = 'same',
                                      kernel_regularizer = self.regularizer, 
                                      kernel_initializer = self.initializer),
               tf.keras.layers.BatchNormalization(),
               tf.keras.layers.LeakyReLU(),
               tf.keras.layers.AveragePooling2D(pool_size=(2, 2),
                         strides=(1, 1), padding='same'),
               tf.keras.layers.Conv2D(filters            = self.filterBase * 4, 
                                      kernel_size        = 3, 
                                      strides            = (1,1),
                                      padding            = 'same',
                                      kernel_regularizer = self.regularizer, 
                                      kernel_initializer = self.initializer),
               tf.keras.layers.BatchNormalization(),
               tf.keras.layers.LeakyReLU(),                         
      ], name="downSampleLayer_256_3_BN_1")
      
      self.downSampleLayer_512_3_BN_0 = tf.keras.Sequential([
               tf.keras.layers.Conv2D(filters            = self.filterBase * 8, 
                                      kernel_size        = 3, 
                                      strides            = (2,2),
                                      padding            = 'same',
                                      kernel_regularizer = self.regularizer, 
                                      kernel_initializer = self.initializer),
               tf.keras.layers.BatchNormalization(),
               tf.keras.layers.LeakyReLU(),          
      ], name="downSampleLayer_512_3_BN_0")
      
      self.downSampleLayer_512_3_BN_1 = tf.keras.Sequential([
               tf.keras.layers.Conv2D(filters            = self.filterBase * 8, 
                                      kernel_size        = 3, 
                                      strides            = (2,2),
                                      padding            = 'same',
                                      kernel_regularizer = self.regularizer, 
                                      kernel_initializer = self.initializer),
               tf.keras.layers.BatchNormalization(),
               tf.keras.layers.LeakyReLU(),          
      ], name="downSampleLayer_512_3_BN_1")
      
      self.downSampleLayer_1024_3_BN_0 = tf.keras.Sequential([
               tf.keras.layers.Conv2D(filters            = self.filterBase * 16, 
                                      kernel_size        = 3, 
                                      strides            = (2,2),
                                      padding            = 'same',
                                      kernel_regularizer = self.regularizer, 
                                      kernel_initializer = self.initializer),
               tf.keras.layers.BatchNormalization(),
               tf.keras.layers.LeakyReLU(),          
      ], name="downSampleLayer_1024_3_BN_0")
      
      self.upSampleLayer_1024_3_DO_0 = tf.keras.Sequential([
        tf.keras.layers.Conv2DTranspose(filters            = self.filterBase * 16, 
                                        kernel_size        = 3, 
                                        strides            = (2,2),
                                        padding            = 'same',
                                        kernel_regularizer = self.regularizer, 
                                        kernel_initializer = self.initializer),
        tf.keras.layers.BatchNormalization(),
        tf.keras.layers.LeakyReLU(),      
      ], name="upSampleLayer_1024_3_DO_0")
      
      self.upSampleLayer_512_3_DO_0 = tf.keras.Sequential([
        tf.keras.layers.Conv2DTranspose(filters            = self.filterBase * 8, 
                                        kernel_size        = 3, 
                                        strides            = (2,2),
                                        padding            = 'same',
                                        kernel_regularizer = self.regularizer, 
                                        kernel_initializer = self.initializer),
        tf.keras.layers.BatchNormalization(),        
        tf.keras.layers.LeakyReLU(),
      ], name="upSampleLayer_512_3_DO_0")
      
      self.upSampleLayer_512_3_DO_1 = tf.keras.Sequential([
        tf.keras.layers.Conv2DTranspose(filters            = self.filterBase * 8, 
                                        kernel_size        = 3, 
                                        strides            = (2,2),
                                        padding            = 'same',
                                        kernel_regularizer = self.regularizer, 
                                        kernel_initializer = self.initializer),
        tf.keras.layers.BatchNormalization(),        
        tf.keras.layers.LeakyReLU(),       
      ], name="upSampleLayer_512_3_DO_1")
      
      self.upSampleLayer_256_3_0 = tf.keras.Sequential([
        tf.keras.layers.Conv2DTranspose(filters            = self.filterBase * 4, 
                                        kernel_size        = 3, 
                                        strides            = (2,2),
                                        padding            = 'same',
                                        kernel_regularizer = self.regularizer, 
                                        kernel_initializer = self.initializer),
        tf.keras.layers.BatchNormalization(),
        tf.keras.layers.LeakyReLU(),
      ], name="upSampleLayer_256_3_0")
      
      self.upSampleLayer_128_3_0 = tf.keras.Sequential([
        tf.keras.layers.Conv2DTranspose(filters            = self.filterBase * 2, 
                                        kernel_size        = 3, 
                                        strides            = (2,2),
                                        padding            = 'same',
                                        kernel_regularizer = self.regularizer, 
                                        kernel_initializer = self.initializer),
        tf.keras.layers.BatchNormalization(),
        tf.keras.layers.LeakyReLU(),
      ], name="upSampleLayer_128_3_0")
      
      self.outputConv     = tf.keras.Sequential([
               tf.keras.layers.Conv2DTranspose(filters = 3, 
                              kernel_size        = (3,3), 
                              strides            = (2,2),
                              padding            = 'same',
                              kernel_regularizer = self.regularizer,
                              kernel_initializer = self.initializer
                              ),
               tf.keras.layers.BatchNormalization(),
               tf.keras.layers.LeakyReLU(),           
      ], name="outputConv")

    def crop(self, x, s):

      _, Hx, Wx, _ = x.shape
      _, Hs, Ws, _ = s.shape
      H = Hx if Hx <= Hs else Hs
      W = Wx if Wx <= Ws else Ws

      return x[:, 0:H, 0:W, :], s[:, 0:H, 0:W, :]

    def test(self, x, s):
      return self.crop(x, s)

    def call(self, input_tensor, training=False):
      
      down_stack= [
        self.downSampleLayer_128_3_BN_0, # (batch_size, H, W, C)
        self.downSampleLayer_256_3_BN_0,
        self.downSampleLayer_256_3_BN_1,
        self.downSampleLayer_512_3_BN_0,
        self.downSampleLayer_512_3_BN_1,
        self.downSampleLayer_1024_3_BN_0, 
      ]

      up_stack = [
        self.upSampleLayer_1024_3_DO_0, 
        self.upSampleLayer_512_3_DO_0,  
        self.upSampleLayer_512_3_DO_1,  
        self.upSampleLayer_256_3_0,     
        self.upSampleLayer_128_3_0,     
      ]

      # Encoder
      x         = self.inputLayer(input_tensor) 

      inputSkip = x
      skips = []
      for down in down_stack:
        x = down(x, training)
        skips.append(x)
      skips = reversed(skips[:-1])

      # Upsampling and establishing the skip connections

      for up, skip in zip(up_stack, skips):
        x = up(x, training)
        _, Hx, Wx, _ = x.shape

        if training is False and Hx is not None:
          skip_resized = tf.image.resize(skip, [tf.dtypes.cast(Hx, tf.int32), tf.dtypes.cast(Wx, tf.int32)])
        else:
          skip_resized = skip
        
        x = tf.keras.layers.Concatenate()([x, skip_resized])

      x   = self.outputConv(x)  

      return x

    def build_graph(self):
      x = tf.keras.Input(shape=(None, None, 3))
      return tf.keras.Model(inputs=[x], outputs=self.call(x))

